/*
 * Copyright (c) 2009 Maxim Bourmistrov <maxim@unixconn.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* trafscrambler - anti-sniffer/IDS NKE for OSX */

#include <mach/mach_types.h>
#include <mach/kmod.h>

#include <sys/systm.h>
#include <sys/kpi_mbuf.h>
#include <sys/kern_control.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/socket.h>

#include <netinet/kpi_ipfilter.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip.h>
#include <string.h>

#include "common.h"

#define MBUF_TAG	0xd33d

#define IPHDR_LEN	sizeof(struct ip)
#define TCPHDR_LEN	sizeof(struct tcphdr)
#define UDPHDR_LEN	sizeof(struct udphdr)

static ipfilter_t	filter_ref;
static kern_ctl_ref	kernctl_ref;

static u_int16_t enable = (ENABLE_NKE|ENABLE_DATA|ENABLE_ZERO
				|ENABLE_CKSUM|ENABLE_SEQ);

static mbuf_tag_id_t gidtag;
static mbuf_t syn_dup = NULL;
static mbuf_t push_dup = NULL;

static void inject_mbuf(mbuf_t m);
static void inject_tcp(int, int, mbuf_t, mbuf_t);
static void inject_udp(mbuf_t);

typedef enum PACKETPROCFLAGS {
	INBOUND_DONE = 1,
	OUTBOUND_DONE
} PACKETPROCFLAGS;

enum {
	FAKE_CKSUM,
	FAKE_SEQ,
	SYN_DECOY,
	ZERO_WIN,
	FAKE_DATA
};

static inline long
getrandom (int min, int max)
{
	return ((random () % (int) (((max) + 1) - (min))) + (min));
}

static int 
check_tag(mbuf_t m, mbuf_tag_id_t id_tag,
		mbuf_tag_type_t tag_type, PACKETPROCFLAGS value)
{
	errno_t status;
	int *tag_ref;
	size_t len;
	
	/* Check whether we have seen this packet before. */
	status = mbuf_tag_find(m, id_tag,
				tag_type, &len, (void**)&tag_ref);
	if ((status == 0) && (*tag_ref == value)
				&& (len == sizeof(value)))
		return 1;
		
	return 0;
}

static errno_t
set_tag(mbuf_t m, mbuf_tag_id_t id_tag,
		mbuf_tag_type_t tag_type, PACKETPROCFLAGS value)
{	
	errno_t status;
	int *tag_ref = NULL;
	size_t len;
	
	/* look for existing tag */
	status = mbuf_tag_find(m, id_tag,
				tag_type, &len, (void*)&tag_ref);
	if (status != 0) {		
		status = mbuf_tag_allocate(m, id_tag, tag_type,
						sizeof(value),
						MBUF_WAITOK,
						(void**)&tag_ref);
		if (status == 0)
			*tag_ref = value;
	}

	return status;
}

static void
inject_mbuf(mbuf_t m)
{
	/* Tag the mbuf so we don't process it in ipv4_output_func_cb() */
	if (set_tag(m, gidtag, MBUF_TAG, OUTBOUND_DONE) != 0) {
#if DEBUG
		printf("%s(): failed to set TAG. mbuf NOT injected. freeing.\n", __FUNCTION__);
#endif
		(void)mbuf_freem(m);
		return;
	}

	(void)ipf_inject_output(m, filter_ref, 0);
	return;
}

static void
inject_udp(mbuf_t origin)
{
	mbuf_t to_inject = NULL, in_chain = NULL;
	char *frame;
	int ret;

	struct udphdr *udp_hdr = NULL;

	if (origin == NULL)
		return;
	
	if ((mbuf_dup(origin, MBUF_WAITOK, &to_inject)) != 0)
		return;

	if ((enable & SHOW_INFO) == SHOW_INFO)
		printf("%s: Injected UDP ", NKE_NAME);

	in_chain = to_inject;
	while (in_chain != NULL &&
		(mbuf_flags(in_chain) & MBUF_EOR) != MBUF_EOR) {

		frame = mbuf_data(in_chain);

		/* overwrite the rest */
		if (mbuf_type(in_chain) == MBUF_TYPE_DATA) {
			
			if ((mbuf_flags(in_chain) & MBUF_PKTHDR) == MBUF_PKTHDR) {
				/* random ip_id */
				frame[4] = random();
				frame[5] = random();

				udp_hdr = (struct udphdr *)(frame + IPHDR_LEN);
				udp_hdr->uh_sum = random();
			}

			for(ret = IPHDR_LEN + UDPHDR_LEN; ret < mbuf_len(in_chain); ret++)
				*((u_int8_t *)frame + ret) = getrandom(0, 255);
		}

		in_chain = mbuf_next(in_chain);
	}

	if ((enable & SHOW_INFO) == SHOW_INFO)
		printf("pkt with FAKE_DATA\n");

	inject_mbuf(to_inject);
	return;
}

static void
inject_tcp(int th_flag, int attck_type, mbuf_t origin, mbuf_t dup)
{
	mbuf_t to_inject = NULL, in_chain = NULL;

	char *frame, *frame_orig;
	int ret;

	struct tcphdr *tcp_hdr = NULL, *tcp_hdr_orig = NULL;
	struct ip *ip_hdr = NULL, *ip_hdr_orig = NULL;

	if (origin == NULL)
		return;
	
	if (dup != NULL) {
		if ((mbuf_dup(dup, MBUF_WAITOK, &to_inject)) != 0)
			return;
	} else {
		if ((mbuf_dup(origin, MBUF_WAITOK, &to_inject)) != 0)
			return;
	}

	frame = mbuf_data(to_inject);
	ip_hdr = (struct ip *)frame;
	tcp_hdr = (struct tcphdr *)(frame + IPHDR_LEN);

	if (dup != NULL) {
		frame_orig = mbuf_data(origin);
		ip_hdr_orig = (struct ip *)frame_orig;
		tcp_hdr_orig = (struct tcphdr *)(frame_orig + IPHDR_LEN);	

		ip_hdr->ip_src = ip_hdr_orig->ip_src;
		ip_hdr->ip_dst = ip_hdr_orig->ip_dst;
		
		tcp_hdr->th_sport = tcp_hdr_orig->th_sport;
		tcp_hdr->th_dport = tcp_hdr_orig->th_dport;
		tcp_hdr->th_seq = tcp_hdr_orig->th_seq;
		tcp_hdr->th_ack = tcp_hdr_orig->th_ack;
	}

	if ((enable & SHOW_INFO) == SHOW_INFO)
		printf("%s: Injected ", NKE_NAME);

	switch(attck_type) {
	case FAKE_DATA: {
			/* We inject bogus data here with
			 * bad TCP checksum or wrong th_seq and th_ack numbers
			 * This is done for each valid TH_PUSH|TH_ACK pkt
			 */
			
			int rand = random() % 5;
			if (rand == 0)
				(void)mbuf_clear_csum_performed(to_inject); /* invalidate cksum */

			/* fill in bogus data */
			in_chain = to_inject;
			while (in_chain != NULL &&
					(mbuf_flags(in_chain) & MBUF_EOR) != MBUF_EOR) {

				frame = mbuf_data(in_chain);

				/* overwrite data stored in the header, if any */
				if ((mbuf_flags(in_chain) & MBUF_PKTHDR) == MBUF_PKTHDR) {
					tcp_hdr = (struct tcphdr *)(frame + IPHDR_LEN);
					if (rand != 0) {
						tcp_hdr->th_seq = random();
						tcp_hdr->th_ack = random();
					}

					for(ret = IPHDR_LEN + (tcp_hdr->th_off * 4);
								ret < mbuf_len(to_inject); ret++)
						*((u_int8_t *)frame + ret) = getrandom(0, 255);
				}

				/* overwrite the rest */
				if (mbuf_type(in_chain) == MBUF_TYPE_DATA) {
					for(ret = 0; ret < mbuf_len(in_chain); ret++)
						*((u_int8_t *)frame + ret) = getrandom(0, 255);
				}

				in_chain = mbuf_next(in_chain);
			}

			if ((enable & SHOW_INFO) == SHOW_INFO) {
				printf("pkt with FAKE_DATA");
				(rand == 0) ? printf("(bad TCP cksum)\n") : printf("(bad TCP seq)\n");
			}
		}
		break;
	case FAKE_CKSUM: {
			
			(void)mbuf_clear_csum_performed(to_inject);
			
			if (th_flag == TH_SYN) {
				tcp_hdr->th_flags = (th_flag|TH_URG);
				
				if ((enable & SHOW_INFO) == SHOW_INFO)
					printf("SYN");
			}
			else if (th_flag == TH_FIN) {
				tcp_hdr->th_flags = (th_flag|TH_ACK);
				tcp_hdr->th_ack = random();

				if ((enable & SHOW_INFO) == SHOW_INFO)
					printf("FIN");
			}
			else if (th_flag == TH_RST) {
				tcp_hdr->th_flags = (th_flag|TH_ACK);
				tcp_hdr->th_ack = random();

				if ((enable & SHOW_INFO) == SHOW_INFO)
					printf("RST");
			}
			else if (th_flag == TH_ACK) {
				tcp_hdr->th_flags = th_flag;
				tcp_hdr->th_ack = random();

				if ((enable & SHOW_INFO) == SHOW_INFO)
					printf("ACK");
			} else {
				if ((enable & SHOW_INFO) == SHOW_INFO) 
					printf("original(th_flag: 0x%02x)", tcp_hdr->th_flags);
			}

			if ((enable & SHOW_INFO) == SHOW_INFO)
				printf(" pkt with wrong CKSUM\n");
		}
		break;
	case FAKE_SEQ: {

			tcp_hdr->th_seq = random();

			if (th_flag != TH_SYN && th_flag != 0xdead)
				tcp_hdr->th_ack = random();

			if (th_flag == TH_SYN) {
				tcp_hdr->th_flags = (th_flag|TH_URG);

				if ((enable & SHOW_INFO) == SHOW_INFO)
					printf("SYN");
			}
			else if (th_flag == TH_FIN) {
				tcp_hdr->th_flags = (th_flag|TH_ACK);

				if ((enable & SHOW_INFO) == SHOW_INFO)
					printf("FIN");
			}
			else if (th_flag == TH_RST) {
				tcp_hdr->th_flags = (th_flag|TH_ACK);

				if ((enable & SHOW_INFO) == SHOW_INFO)
					printf("RST");
			}
			else if (th_flag == TH_ACK) {
				tcp_hdr->th_flags = th_flag;

				if ((enable & SHOW_INFO) == SHOW_INFO)
					printf("ACK");
			} else {
				if ((enable & SHOW_INFO) == SHOW_INFO)
					printf("original(th_flag: 0x%02x)", tcp_hdr->th_flags);
			}

			if ((enable & SHOW_INFO) == SHOW_INFO)
				printf(" pkt with wrong SEQ\n");
		}
		break;
	case ZERO_WIN: {
			tcp_hdr->th_seq = random();
			tcp_hdr->th_ack = random();
			tcp_hdr->th_flags = (TH_FIN|TH_SYN|TH_RST|TH_ACK|TH_URG);
			tcp_hdr->th_win = 0;

			if ((enable & SHOW_INFO) == SHOW_INFO)
				printf("pkt with ZERO window\n");
			}
			break;
	case SYN_DECOY: {

			ip_hdr->ip_ttl = getrandom(10, 30);

			/* src IP */
			frame[12] = getrandom(100, 200);
			frame[13] = getrandom(100, 200);
			frame[14] = getrandom(100, 200);
			frame[15] = getrandom(100, 200);

			if ((random() % 5) == 0) {
				tcp_hdr->th_flags |= TH_ACK;
				tcp_hdr->th_ack = random();
			}

			if ((enable & SHOW_INFO) == SHOW_INFO)
				printf("%d spoofed fake SYN(s)\n", NUM_FAKE_SYN);
		}
		break;
	default:
		return;
	}

	/* random ip_id */
	frame[4] = random();
	frame[5] = random();

	inject_mbuf(to_inject);
	return;
}

static errno_t
ipv4_output_func_cb(void* cookie, mbuf_t *data, ipf_pktopts_t options)
{
	mbuf_t m = *data;
	size_t m_len;
	char *frame = NULL;
	int i;

	if ((enable & ENABLE_NKE) != ENABLE_NKE)
		return 0;

	/* Don't process already seen mbuf */
	if (check_tag(m, gidtag, MBUF_TAG, OUTBOUND_DONE)) {
#if DEBUG
		printf("%s(): mbuf already seen\n", __FUNCTION__);
#endif
		return 0;
	}

	/* no use of mbuf we can't tag */
	if ((mbuf_flags(m) & MBUF_PKTHDR) != MBUF_PKTHDR)
		return 0;

	frame = mbuf_data(m);
	m_len = mbuf_len(m);

	if (frame == NULL)
		return 0;	

	if (m_len >= (IPHDR_LEN + TCPHDR_LEN) &&
			(u_int8_t)frame[9] == IPPROTO_TCP) {

		switch((u_int8_t)(frame[33] & ~(TH_URG|TH_ECE|TH_CWR))) {
			case TH_SYN:
			{
#if DEBUG
				printf("Got SYN\n");
#endif
				/* catch an empty pkt for later use */
				if (syn_dup == NULL)
					mbuf_dup(m, MBUF_WAITOK, &syn_dup);

				if ((enable & ENABLE_DECOY) == ENABLE_DECOY)
					/* Inject number of fake SYN pkts */
					for (i = 0; i < NUM_FAKE_SYN; i++)
						(void)inject_tcp(0xdead, SYN_DECOY, m, NULL);

				if ((enable & ENABLE_ZERO) == ENABLE_ZERO)
					(void)inject_tcp(0xdead, ZERO_WIN, m, NULL);

				if ((enable & ENABLE_SEQ) == ENABLE_SEQ) {
					(void)inject_tcp(TH_ACK, FAKE_SEQ, m, NULL);
					(void)inject_tcp(TH_RST, FAKE_SEQ, m, NULL);
				}

				/* Pre-connection SYN */
				if ((enable & ENABLE_CKSUM) == ENABLE_CKSUM)
					(void)inject_tcp(0xdead, FAKE_CKSUM, m, NULL);
				
			}
			break;
			case TH_ACK:
			{
#if DEBUG
				printf("Got ACK\n");
#endif
				if ((enable & ENABLE_DATA) == ENABLE_DATA)
					if (push_dup != NULL)
						(void)inject_tcp(0xdead, FAKE_DATA, m, push_dup);

				if ((random() % 5) == 0) {
					if ((enable & ENABLE_SEQ) == ENABLE_SEQ) {
						(void)inject_tcp(TH_SYN, FAKE_SEQ, m, NULL);
						(void)inject_tcp(TH_RST, FAKE_SEQ, m, NULL);
					}

					if ((enable & ENABLE_ZERO) == ENABLE_ZERO)
						(void)inject_tcp(0xdead, ZERO_WIN, m, NULL);

					if ((enable & ENABLE_CKSUM) == ENABLE_CKSUM)
						(void)inject_tcp(0xdead, FAKE_CKSUM, m, NULL);
				}
			}
			break;
			case (TH_PUSH|TH_ACK):
			{
#if DEBUG
				printf("Got PUSH/ACK\n");
#endif
				/* catch pkt for later use.
				 * size of injected data pkts will be in the same
				 * as the catched one.
				 */
				if (push_dup == NULL)
					mbuf_dup(m, MBUF_WAITOK, &push_dup);

				if ((enable & ENABLE_SEQ) == ENABLE_SEQ) {
					if (syn_dup != NULL) {

						/* TCP reset attack
				 	 	 * http://kerneltrap.org/node/3072
				 	 	 */
						(void)inject_tcp(TH_RST, FAKE_SEQ, m, syn_dup);

						(void)inject_tcp(TH_ACK, FAKE_SEQ, m, syn_dup);

						/* Post-connection SYN */			
						(void)inject_tcp(TH_SYN, FAKE_SEQ, m, syn_dup);
					}
				}

				if ((enable & ENABLE_DATA) == ENABLE_DATA)
					(void)inject_tcp(0xdead, FAKE_DATA, m, NULL);
			}
			break;
			case (TH_FIN|TH_ACK):
			{
#if DEBUG
				printf("Got FIN/ACK\n");
#endif
				if ((enable & ENABLE_DATA) == ENABLE_DATA)
					if (push_dup != NULL)
						(void)inject_tcp(0xdead, FAKE_DATA, m, push_dup);

				if ((enable & ENABLE_SEQ) == ENABLE_SEQ)
					(void)inject_tcp(TH_RST, FAKE_SEQ, m, NULL);

				if ((enable & ENABLE_ZERO) == ENABLE_ZERO)
					(void)inject_tcp(0xdead, ZERO_WIN, m, NULL);

				if ((enable & ENABLE_CKSUM) == ENABLE_CKSUM)
					(void)inject_tcp(0xdead, FAKE_CKSUM, m, NULL);

				if ((enable & ENABLE_DECOY) == ENABLE_DECOY)
					for (i = 0; i < NUM_FAKE_SYN; i++)
						(void)inject_tcp(0xdead, SYN_DECOY, m, NULL);

			}
			break;
			case (TH_RST|TH_ACK):
			{
#if DEBUG
				printf("Got RST/ACK\n");
#endif
				if ((enable & ENABLE_DATA) == ENABLE_DATA)
					if (push_dup != NULL)
						(void)inject_tcp(0xdead, FAKE_DATA, m, push_dup);

				if ((enable & ENABLE_SEQ) == ENABLE_SEQ)
					(void)inject_tcp(TH_FIN, FAKE_SEQ, m, NULL);

				if ((enable & ENABLE_ZERO) == ENABLE_ZERO)
					(void)inject_tcp(0xdead, ZERO_WIN, m, NULL);

				if ((enable & ENABLE_CKSUM) == ENABLE_CKSUM)
					(void)inject_tcp(0xdead, FAKE_CKSUM, m, NULL);

				if ((enable & ENABLE_DECOY) == ENABLE_DECOY)
					for (i = 0; i < NUM_FAKE_SYN; i++)
						(void)inject_tcp(0xdead, SYN_DECOY, m, NULL);

			}
			break;
			case (TH_SYN|TH_ACK):
			case TH_FIN:
			case TH_RST:
			default: break;
		}
	}

	if (m_len >= (IPHDR_LEN + UDPHDR_LEN) &&
			(u_int8_t)frame[9] == IPPROTO_UDP) {
		/* (void)inject_udp(m); */
	}

	return 0;
}

/* -------- NKE control functions ------------------------ */
errno_t
ts_ctl_conn(kern_ctl_ref kctlref,
		struct sockaddr_ctl *sac, void **unitinfo)
{
	return 0;
}

errno_t
ts_ctl_disconn(kern_ctl_ref kctlref,
		u_int32_t unit, void *unitinfo)
{
	return 0;
}

errno_t
ts_ctl_setopt(kern_ctl_ref kctlref, u_int32_t unit,
			void *unitinfo, int opt,
				void *data, size_t *len)
{
	char *out = "";

	switch(opt) {
		case ENABLE_NKE:
			enable |= opt;
			out = "enabled";
			break;
		case DISABLE_NKE:
			enable &= opt;
			out = "disabled";	
			break;
		case ENABLE_DECOY:
			enable |= opt;
			out = "enabled SYN decoy injection";
			break;
		case DISABLE_DECOY:
			enable &= opt;
			out = "disabled SYN decoy injection";
			break;
		case ENABLE_DATA:
			enable |= opt;
			out = "enabled fake data injection";
			break;
		case DISABLE_DATA:
			enable &= opt;
			out = "disabled fake data injection";
			break;
		case ENABLE_ZERO:
			enable |= opt;
			out = "enabled zero win injection";
			break;
		case DISABLE_ZERO:
			enable &= opt;
			out = "disabled zero win injection";
			break;
		case ENABLE_CKSUM:
			enable |= opt;
			out = "enabled bad cksum injection";
			break;
		case DISABLE_CKSUM:
			enable &= opt;
			out = "disabled bad cksum injection";
			break;
		case ENABLE_SEQ:
			enable |= opt;
			out = "enabled bad seq injection";
			break;
		case DISABLE_SEQ:
			enable &= opt;
			out = "disabled bad seq injection";
			break;
		case SHOW_INFO:
			enable |= opt;
			out = "enabled info output";
			break;
		case HIDE_INFO:
			enable &= opt;
			out = "disabled info output";
			break;
		case ENABLE_ALL:
			enable |= opt;
			out = "enabled ALL options";
			break;
		case DISABLE_ALL:
			enable &= ~(ENABLE_ALL);
			out = "disabled ALL options";
			break;
		default: return EINVAL;
	}

	printf("%s(%s) successfully %s\n", NKE_NAME, NKE_VERSION, out);
	return 0;
}

errno_t
ts_ctl_getopt(kern_ctl_ref kctlref, u_int32_t unit,
			void *unitinfo, int opt,
				void *data, size_t *len)
{
	return 0;
}

/* ----- NKE init --------------------------------- */

static struct ipf_filter ipv4_filter = {
	.cookie = NULL,
	.name = NKE_NAME,
	.ipf_input = NULL,
	.ipf_output = ipv4_output_func_cb,
	.ipf_detach = NULL,
};

static struct kern_ctl_reg ts_ctl = {
	.ctl_name = NKE_NAME,
	.ctl_id = 0,
	.ctl_unit = 0,
	.ctl_flags = (CTL_FLAG_PRIVILEGED & CTL_FLAG_REG_ID_UNIT),
	.ctl_connect = ts_ctl_conn,
	.ctl_disconnect = ts_ctl_disconn,
	.ctl_send = NULL,
	.ctl_setopt = ts_ctl_setopt,
	.ctl_getopt = ts_ctl_getopt,
};

static kern_return_t
trafscrambler_module_start(struct kmod_info *ki, void *data)
{
	int ret;

	ret = mbuf_tag_id_find(NKE_NAME, &gidtag);
	if (ret != 0) {
		printf("mbuf_tag_id_find returned error %d\n", ret);
		goto out;
	}
	
	ret = ipf_addv4(&ipv4_filter, &filter_ref);
	if (ret != 0)
		goto out;

	ret = ctl_register(&ts_ctl, &kernctl_ref);

	printf("%s(%s) was successfully loaded!\n", NKE_NAME, NKE_VERSION);
	return KERN_SUCCESS;

out:
	printf("%s(%s) failed to load!\n", NKE_NAME, NKE_VERSION);
	return KERN_FAILURE;
}

static kern_return_t
trafscrambler_module_stop(struct kmod_info *ki, void *data)
{
	errno_t ret;

	ret = ctl_deregister(kernctl_ref);
	if (ret != 0)
		printf("%s(%s): Failed to deregister ts_ctl\n", NKE_NAME, NKE_VERSION);

	ret = ipf_remove(filter_ref);
	if (ret != 0)
		printf("%s(%s): Failed uninstall ipfiler\n", NKE_NAME, NKE_VERSION);

	if (syn_dup != NULL)
		(void)mbuf_freem(syn_dup);

	if (push_dup != NULL)
		(void)mbuf_freem(push_dup);

	printf("%s(%s) was succsessfully unloaded!\n", NKE_NAME, NKE_VERSION);
	return KERN_SUCCESS;
}

KMOD_DECL(trafscrambler, NKE_VERSION);

